from abc import abstractmethod
from typing import AsyncIterator, Optional

from langchain_core.messages import AIMessageChunk
from langchain_core.runnables import RunnableConfig
from langchain_core.tracers.context import collect_runs
from langgraph.prebuilt.tool_executor import ToolExecutor, ToolInvocation

from docugami_langchain.agents.models import (
    AgentState,
    CitedAnswer,
    Invocation,
    StepState,
)
from docugami_langchain.base_runnable import BaseRunnable, TracedResponse
from docugami_langchain.tools.common import BaseDocugamiTool

THINKING = "Thinking..."


class BaseDocugamiAgent(BaseRunnable[AgentState]):
    """
    Base class with common functionality for various chains.
    """

    tools: list[BaseDocugamiTool] = []

    @abstractmethod
    def parse_final_answer(self, text: str) -> str: ...

    def execute_tool(
        self,
        state: AgentState,
        config: Optional[RunnableConfig],
    ) -> AgentState:
        """
        Gets the most recent tool invocation (added by the agent) and execute it.
        """

        inv_model = state.get("tool_invocation")
        if not inv_model:
            raise Exception(f"No tool invocation in model: {state}")

        previous_steps = state.get("intermediate_steps") or []
        if previous_steps and any(
            [s for s in previous_steps if s.invocation == inv_model]
        ):
            tool_output = CitedAnswer(
                source=inv_model.tool_name,
                answer="This tool has been invoked before with identical inputs. Please try different inputs or a different tool, after reconsidering previous thoughts and observations. "
                + "Be careful you don't get stuck in a loop.",
            )
        else:
            tool_executor = ToolExecutor(self.tools)
            tool_output = tool_executor.invoke(
                ToolInvocation(  # LangChain version of the Invocation object
                    tool=inv_model.tool_name,
                    tool_input=inv_model.tool_input,
                ),
                config,
            )
            if not isinstance(tool_output, CitedAnswer):
                raise Exception(f"Invalid output from tool executor: {tool_output}")

        step = StepState(
            invocation=inv_model,
            output=tool_output.answer,
            citations=tool_output.citations,
        )
        return {"intermediate_steps": previous_steps + [step]}

    def invocation_answer(
        self,
        invocation: Invocation,
        answer_source: str,
    ) -> AgentState:
        """
        Builds a human readable interim answer from a tool invocation.
        """

        tool_name = invocation.tool_name
        tool_input = invocation.tool_input
        if tool_name and tool_input:
            busy_text = THINKING
            match = [t for t in self.tools if t.name.lower() == tool_name]
            if match:
                busy_text = match[0].to_human_readable(invocation)

        return {
            "tool_invocation": invocation,
            "cited_answer": CitedAnswer(
                source=answer_source,
                answer=busy_text,  # Show the user interim output.
            ),
        }

    def run(  # type: ignore[override]
        self,
        question: str,
        chat_history: list[tuple[str, str]] = [],
        config: Optional[RunnableConfig] = None,
    ) -> TracedResponse[AgentState]:
        if not question:
            raise Exception("Input required: question")

        return super().run(
            question=question,
            chat_history=chat_history,
            config=config,
        )

    def run_batch(  # type: ignore[override]
        self,
        inputs: list[tuple[str, list[tuple[str, str]]]],
        config: Optional[RunnableConfig] = None,
    ) -> list[AgentState]:
        return super().run_batch(
            inputs=[
                {
                    "question": i[0],
                    "chat_history": i[1],
                }
                for i in inputs
            ],
            config=config,
        )

    async def run_stream(  # type: ignore[override]
        self,
        question: str,
        chat_history: list[tuple[str, str]] = [],
        config: Optional[RunnableConfig] = None,
    ) -> AsyncIterator[TracedResponse[AgentState]]:
        if not question:
            raise Exception("Input required: question")

        config, kwargs_dict = self._prepare_run_args(
            {
                "question": question,
                "chat_history": chat_history,
            }
        )

        with collect_runs() as cb:
            last_response_value = None
            current_step_token_stream = ""
            final_streaming_started = False
            async for output in self.runnable().astream_log(
                input=kwargs_dict,
                config=config,
                include_types=["llm"],
            ):
                for op in output.ops:
                    op_path = op.get("path", "")
                    op_value = op.get("value", "")
                    if not final_streaming_started and op_path == "/streamed_output/-":
                        # Restart token stream for each interim step
                        current_step_token_stream = ""
                        if not isinstance(op_value, dict):
                            # Agent step-wise streaming yields dictionaries keyed by node name
                            # Ref: https://python.langchain.com/docs/langgraph#streaming-node-output
                            raise Exception(
                                "Expected dictionary output from agent streaming"
                            )

                        if not len(op_value.keys()) == 1:
                            raise Exception(
                                "Expected output from one node at a time in step-wise agent streaming output"
                            )

                        key = list(op_value.keys())[0]
                        last_response_value = op_value[key]
                        yield TracedResponse[AgentState](value=last_response_value)
                    elif op_path.startswith("/logs/") and op_path.endswith(
                        "/streamed_output/-"
                    ):
                        # Because we chose to only include LLMs, these are LLM tokens
                        if isinstance(op_value, AIMessageChunk):
                            current_step_token_stream += str(op_value.content)

                            final_answer = self.parse_final_answer(
                                current_step_token_stream
                            )

                            if final_answer:
                                if not final_streaming_started:
                                    # Set final streaming started once as soon as we see the final
                                    # answer action in the token stream
                                    final_streaming_started = bool(final_answer)
                                else:
                                    # Start streaming the final answer, no more interim steps
                                    last_response_value = AgentState(
                                        chat_history=[],
                                        question="",
                                        tool_invocation=None,
                                        intermediate_steps=[],
                                        cited_answer=CitedAnswer(
                                            source=self.__class__.__name__,
                                            answer=final_answer,
                                            is_final=True,
                                        ),
                                    )
                                    yield TracedResponse[AgentState](
                                        value=last_response_value
                                    )

            # Yield the final result with the run_id
            if last_response_value:
                if "cited_answer" in last_response_value:
                    last_response_value["cited_answer"].is_final = True

            if cb.traced_runs:
                run_id = str(cb.traced_runs[0].id)
                yield TracedResponse[AgentState](
                    run_id=run_id,
                    value=last_response_value,  # type: ignore
                )



from docugami_langchain.chains.querying.docugami_explained_sql_query_chain import (
    DocugamiExplainedSQLQueryChain,
)
from docugami_langchain.chains.querying.sql_fixup_chain import SQLFixupChain
from docugami_langchain.chains.querying.sql_query_explainer_chain import (
    SQLQueryExplainerChain,
)
from docugami_langchain.chains.querying.sql_result_chain import SQLResultChain
from docugami_langchain.chains.querying.sql_result_explainer_chain import (
    SQLResultExplainerChain,
)

__all__ = [
    "DocugamiExplainedSQLQueryChain",
    "SQLFixupChain",
    "SQLQueryExplainerChain",
    "SQLResultChain",
    "SQLResultExplainerChain",
]



from docugami_langchain.chains.chunks.elaborate_chunk_chain import ElaborateChunkChain
from docugami_langchain.chains.chunks.summarize_chunk_chain import SummarizeChunkChain

__all__ = [
    "SummarizeChunkChain",
    "ElaborateChunkChain",
]



from typing import AsyncIterator, Optional

from langchain_core.runnables import RunnableConfig

from docugami_langchain.base_runnable import TracedResponse
from docugami_langchain.chains.base import BaseDocugamiChain
from docugami_langchain.params import RunnableParameters, RunnableSingleParameter


class SQLResultExplainerChain(BaseDocugamiChain[str]):
    def params(self) -> RunnableParameters:
        return RunnableParameters(
            inputs=[
                RunnableSingleParameter(
                    "sql_query",
                    "SQL QUERY",
                    "SQL Query that was run by the system, to answer the question asked.",
                ),
                RunnableSingleParameter(
                    "sql_result",
                    "SQL RESULT",
                    "Result of the SQL Query.",
                ),
                RunnableSingleParameter(
                    "question",
                    "QUESTION",
                    "Question asked by the user.",
                ),
            ],
            output=RunnableSingleParameter(
                "answer",
                "ANSWER",
                "Human readable answer based on the question, SQL Query and the SQL Result, considering the rules and examples provided. Please give a short one line answer, only "
                + "describing the result and not the query. Remember not to mention SQL or tables as instructed, just describe the result.",
            ),
            task_description="acts as a SQLite expert and given an input SQL Query, a SQL Result and a question, creates a human readable answer based on the SQL Result",
            additional_instructions=[
                "- Shorter answers are better, but make sure you always use all the data in the SQL Result.",
                "- In your answer, never mention SQL or the fact that you are producing a human readable result based on SQL results.",
            ],
            key_finding_output_parse=False,  # set to False for streaming
        )

    def run(  # type: ignore[override]
        self,
        question: str,
        sql_query: str,
        sql_result: str,
        config: Optional[RunnableConfig] = None,
    ) -> TracedResponse[str]:
        if not question or not sql_query:
            raise Exception("Inputs required: question, sql_query")

        return super().run(
            question=question,
            sql_query=sql_query,
            sql_result=sql_result,
            config=config,
        )

    async def run_stream(  # type: ignore[override]
        self,
        question: str,
        sql_query: str,
        sql_result: str,
        config: Optional[RunnableConfig] = None,
    ) -> AsyncIterator[TracedResponse[str]]:
        if not question or not sql_query:
            raise Exception("Inputs required: question, sql_query")

        async for item in super().run_stream(
            question=question,
            sql_query=sql_query,
            sql_result=sql_result,
            config=config,
        ):
            yield item

    def run_batch(  # type: ignore[override]
        self,
        inputs: list[tuple[str, str, str]],
        config: Optional[RunnableConfig] = None,
    ) -> list[str]:
        return super().run_batch(
            inputs=[
                {
                    "question": i[0],
                    "sql_query": i[1],
                    "sql_result": i[2],
                }
                for i in inputs
            ],
            config=config,
        )



from operator import itemgetter
from typing import AsyncIterator, Optional

from langchain_core.documents import Document
from langchain_core.runnables import Runnable, RunnableConfig, RunnableLambda

from docugami_langchain.base_runnable import TracedResponse
from docugami_langchain.chains.base import BaseDocugamiChain
from docugami_langchain.history import chat_history_to_str
from docugami_langchain.output_parsers.line_separated_list import (
    LineSeparatedListOutputParser,
)
from docugami_langchain.params import RunnableParameters, RunnableSingleParameter
from docugami_langchain.utils.documents import formatted_summaries


class SuggestedReportChain(BaseDocugamiChain[list[str]]):
    def runnable(self) -> Runnable:
        """
        Custom runnable for this chain.
        """

        return {
            "summaries": itemgetter("summaries") | RunnableLambda(formatted_summaries),
            "chat_history": itemgetter("chat_history"),
        } | super().runnable()

    def params(self) -> RunnableParameters:
        return RunnableParameters(
            inputs=[
                RunnableSingleParameter(
                    "summaries",
                    "SUMMARIES",
                    "Summaries of representative documents from a set of documents.",
                ),
                RunnableSingleParameter(
                    "chat_history",
                    "CHAT HISTORY",
                    "Previous chat messages that may provide additional context for the type of columns that should be generated.",
                ),
            ],
            output=RunnableSingleParameter(
                "suggested_report_columns",
                "SUGGESTED REPORT COLUMNS",
                "Suggested columns for an automatically generated report against documents similar to the ones provided.",
            ),
            task_description="suggests columns for an automatically generated report against a set of documents, given some summaries of representative documents from the set",
            additional_instructions=[
                "- Generate 'human-like' column labels, i.e. things a human familiar with this particular set of documents might want to know in a diagnostic report about this set of documents",
                "- Bias towards columns highly likely to be found in all or most of the documents.",
                "- Avoid columns that are highly likely to contain boilerplate or uninteresting information that is similar for all the documents.",
                "- Do not include Document Name or File Name in your list, since those are included automatically by the system.",
                "- Make sure the column names you generate are alphanumeric, containing no special characters or parentheses.",
                "- Generate up to 20 suggested columns as a list, one per line.",
            ],
            additional_runnables=[LineSeparatedListOutputParser()],
            stop_sequences=["\n\n", "<|eot_id|>"],
            key_finding_output_parse=False,  # set to False for streaming
        )

    def run(  # type: ignore[override]
        self,
        summaries: list[Document],
        chat_history: list[tuple[str, str]] = [],
        config: Optional[RunnableConfig] = None,
    ) -> TracedResponse[list[str]]:
        if not summaries:
            raise Exception("Input required: summaries")

        return super().run(
            summaries=summaries,
            chat_history=chat_history_to_str(chat_history),
            config=config,
        )

    async def run_stream(  # type: ignore[override]
        self,
        summaries: list[Document],
        chat_history: list[tuple[str, str]] = [],
        config: Optional[RunnableConfig] = None,
    ) -> AsyncIterator[TracedResponse[list[str]]]:
        if not summaries:
            raise Exception("Input required: summaries")

        async for item in super().run_stream(
            summaries=summaries,
            chat_history=chat_history_to_str(chat_history),
            config=config,
        ):
            yield item

    def run_batch(  # type: ignore[override]
        self,
    ) -> list[list[str]]:
        raise NotImplementedError()



import json
import re
from typing import Union

from langchain_core.exceptions import OutputParserException
from langchain_core.output_parsers import BaseOutputParser

from docugami_langchain.agents.models import Invocation
from docugami_langchain.utils.string_cleanup import clean_text

THOUGHT_MARKER = "Thought:"
OBSERVATION_MARKER = "Observation:"
FINAL_ANSWER_MARKER = "Final Answer:"

STRICT_REACT_PATTERN = re.compile(r"^.*?`{3}(?:json)?\n?(.*?)`{3}.*?$", re.DOTALL)
"""Regex pattern to parse the output strictly, JSON delimited by ``` as instructed in a ReAct prompt."""

SIMPLE_JSON_PATTERN = re.compile(r"(\{[^}]*\})")
"""Regex pattern to just find any simple (non-nested) JSON in the output, not delimited by anything."""


class CustomReActJsonSingleInputOutputParser(BaseOutputParser[Union[Invocation, str]]):
    """
    A custom version of ReActJsonSingleInputOutputParser from the
    langchain lib with the following changes:

    1. Decouples from langchain dependency and returns a simple custom TypedDict.
    2. If the standard ReAct style output is not found in the text, try to parse
    any json found in the text and return that if it matches the return type.
    3. Permissive parsing mode that assumes unparsable output is final answer,
    since some weaker models fail to respect the ReAct prompt format when producing
    the final answer.

    Ref: libs/langchain/langchain/agents/output_parsers/react_json_single_input.py
    """

    permissive = True
    """Softer parsing. Specifies whether unparsable input is considered final output."""

    @property
    def _type(self) -> str:
        return "custom-react-json-single-input"

    def _parse_regex(self, text: str, regex: re.Pattern[str]) -> dict:
        found = regex.search(text)
        if not found:
            raise ValueError("Invocation text not found")
        invocation_text = found.group(1)
        invocation_text = clean_text(invocation_text, protect_nested_strings=True)

        return json.loads(invocation_text.strip())

    def parse(self, text: str) -> Union[Invocation, str]:
        includes_answer = FINAL_ANSWER_MARKER in text

        try:
            # First, try parsing with STRICT_REACT_PATTERN
            response = self._parse_regex(text, STRICT_REACT_PATTERN)
            tool_name = response.get("tool_name", "")

            if not tool_name:
                raise Exception(f"could not find tool_name in text: {text}")

            return Invocation(
                tool_name=tool_name,
                tool_input=response.get("tool_input", ""),
                log=text,
            )
        except Exception:
            # Next, try parsing with SIMPLE_JSON_PATTERN
            try:
                response = self._parse_regex(text, SIMPLE_JSON_PATTERN)
                tool_name = response.get("tool_name", "")

                if not tool_name:
                    raise Exception(f"could not find tool_name in text: {text}")

                return Invocation(
                    tool_name=tool_name,
                    tool_input=response.get("tool_input", ""),
                    log=text,
                )
            except Exception:
                # If neither pattern matches, handle according to permissive mode
                if not includes_answer:
                    if not self.permissive:
                        raise OutputParserException(
                            f"Could not parse LLM output: {text}"
                        )

                output = text.split(FINAL_ANSWER_MARKER)[-1].strip()

                if "{" in output:
                    raise OutputParserException(
                        f"Potential JSON in parsed output {output}"
                    )

                if THOUGHT_MARKER in output:
                    raise OutputParserException(
                        f"Potential Thought Marker in parsed output {output}"
                    )

                if OBSERVATION_MARKER in output:
                    raise OutputParserException(
                        f"Potential Observation Marker in parsed output {output}"
                    )

                output = clean_text(output)

                return output



from typing import TypedDict


class ExplainedSQLResult(TypedDict):
    sql_query: str
    explained_sql_query: str
    sql_result: str
    explained_sql_result: str


class ExplainedSQLQuestionResult(TypedDict):
    question: str
    result: ExplainedSQLResult



from __future__ import annotations

import operator
from typing import Annotated, Sequence, TypedDict, Union

from langchain_core.pydantic_v1 import BaseModel


class Citation(BaseModel):
    label: str
    details: str = ""
    link: str = ""


class CitedAnswer(BaseModel):
    source: str
    answer: str
    citations: list[Citation] = []
    is_final: bool = False
    metadata: dict = {}


class Invocation(BaseModel):
    tool_name: str
    tool_input: str
    log: str = ""

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Invocation):
            return NotImplemented

        # Compare tool_name and tool_input for equality
        return (self.tool_name, self.tool_input) == (other.tool_name, other.tool_input)


class StepState(BaseModel):
    output: str
    invocation: Invocation
    citations: list[Citation] = []

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, StepState):
            return NotImplemented

        # Compare invocation and output for equality
        return (self.invocation, self.output) == (other.invocation, other.output)


class AgentState(TypedDict):
    # **** Inputs
    # The list of previous messages in the conversation
    chat_history: list[tuple[str, str]]

    # The input question
    question: str

    # **** Internal State

    # Names of all tools available to the agent
    tool_names: Union[str, None]

    # Descriptions of all tools available to the agent
    tool_descriptions: Union[str, None]

    # The next tool invocation that must be made
    tool_invocation: Union[Invocation, None]

    # List of steps taken so far (this state is added to, not overwritten)
    intermediate_steps: Annotated[Sequence[StepState], operator.add]

    # **** Output
    cited_answer: CitedAnswer



from abc import abstractmethod
from typing import Any, AsyncIterator

from langchain_core.runnables.utils import AddableDict
from langchain_core.tracers.context import collect_runs

from docugami_langchain.base_runnable import BaseRunnable, T, TracedResponse


class BaseDocugamiChain(BaseRunnable[T]):
    """
    Base class with common functionality for various chains.
    """

    @abstractmethod
    async def run_stream(self, **kwargs: Any) -> AsyncIterator[TracedResponse[T]]:  # type: ignore[override, misc]
        config, kwargs_dict = self._prepare_run_args(kwargs)

        with collect_runs() as cb:
            incremental_answer = None
            async for chunk in self.runnable().astream(
                input=kwargs_dict,
                config=config,  # type: ignore
            ):
                if isinstance(chunk, dict):
                    chunk = AddableDict(chunk)

                if not incremental_answer:
                    incremental_answer = chunk
                else:
                    incremental_answer += chunk

                yield TracedResponse[T](value=incremental_answer)

            # yield the final result with the run_id
            if cb.traced_runs:
                run_id = str(cb.traced_runs[0].id)
                yield TracedResponse[T](
                    run_id=run_id,
                    value=incremental_answer,  # type: ignore
                )



import re
import string

UN_ESCAPE_MAP = {
    "\\\\": "\\",
    "\\'": "'",
    '\\"': '"',
    "\\b": "\b",
    "\\f": "\f",
    "\\n": "\n",
    "\\r": "\r",
    "\\t": "\t",
    "\\(": "(",
    "\\)": ")",
    "\\[": "[",
    "\\]": "]",
    "\\_": "_",
}


def _replace_null_outside_quotes(text: str) -> str:
    """
    Looks for null outside quotes, and if found replaces it with "".
    """

    text = text.strip()
    if not text:
        return ""

    def replacement(match: re.Match) -> str:
        before = text[: match.start()]
        if before.count('"') % 2 == 0:  # Even number of quotes before 'null'
            return '""'
        else:
            return str(match.group(0))  # 'null' is inside quotes, don't replace

    return re.sub(r"\bnull\b", replacement, text, flags=re.IGNORECASE)


def _escape_non_escape_sequence_backslashes(text: str) -> str:
    """
    Escape backslashes that are not part of a known escape sequence.

    Looks for a backslash that is not a part of any known escape characters ('n', 'r', 't', 'f', '"', "'", '\\'), and escapes it.
    """
    return re.sub(r'\\(?![nrtf"\'\\])', r"\\\\", text)


def _unescape_escaped_chars_outside_quoted_strings(text: str) -> str:
    """
    Unescapes unnecessary escaped characters outside of quoted strings, e.g., in a SQL query.

    Assumes that a quoted string starts and ends with the same type of quote
    (single or double) and does not contain mixed types.
    """
    text = text.strip()
    if not text:
        return ""

    def replacement(match: re.Match) -> str:
        before = text[: match.start()]
        # Count the occurrences of unescaped single and double quotes before the match
        single_quotes_count = len(re.findall(r"(?<!\\)'", before))
        double_quotes_count = len(re.findall(r'(?<!\\)"', before))

        matched_sequence = match.group(0)
        # Determine if the match is outside quotes based on the counts
        if single_quotes_count % 2 == 0 and double_quotes_count % 2 == 0:
            # If both counts are even, we are outside quotes, so unescape
            if matched_sequence in UN_ESCAPE_MAP:
                return UN_ESCAPE_MAP[matched_sequence]
            else:
                return matched_sequence[1:]  # Skip the backslash to unescape
        else:
            # Inside quotes, keep the original escaped character
            return matched_sequence

    # This regex looks for any escaped character
    return re.sub(r"\\(.)", replacement, text)


def _unescaped_all_escape_sequences(text: str) -> str:
    """
    Unescapes over-escaped text, i.e. where valid escape sequences are escaped.
    """

    text = text.strip()
    if not text:
        return ""

    for sequence in UN_ESCAPE_MAP:
        text = text.replace(sequence, UN_ESCAPE_MAP[sequence])

    return text


def clean_text(text: str, protect_nested_strings: bool = False) -> str:
    """
    Cleans the given text.
    """

    text = text.strip()
    if not text:
        return ""

    text = "".join(filter(lambda x: x in string.printable, text))  # non-printable chars
    text = _replace_null_outside_quotes(text)

    if protect_nested_strings:
        text = _escape_non_escape_sequence_backslashes(text)
        text = _unescape_escaped_chars_outside_quoted_strings(text)
    else:
        text = _unescaped_all_escape_sequences(text)

    return text



from datetime import datetime
from typing import Any, AsyncIterator, Optional

from docugami_langchain.base_runnable import TracedResponse
from docugami_langchain.chains.base import BaseDocugamiChain
from docugami_langchain.output_parsers.datetime import DatetimeOutputParser
from docugami_langchain.params import RunnableParameters, RunnableSingleParameter

OUTPUT_FORMAT = "%m/%d/%Y"


class DateParseChain(BaseDocugamiChain[datetime]):
    def params(self) -> RunnableParameters:
        return RunnableParameters(
            inputs=[
                RunnableSingleParameter(
                    "date_text",
                    "DATE TEXT",
                    "The date expression that needs to be parsed, in rough natural language with possible typos or OCR glitches.",
                ),
            ],
            output=RunnableSingleParameter(
                "parsed_date",
                "PARSED DATE",
                f"The result of parsing the date expression, in {OUTPUT_FORMAT} format.",
            ),
            task_description=f"parses date expressions specified in rough natural language, producing output strictly in the standard {OUTPUT_FORMAT} format",
            additional_instructions=[
                f"- Always produce output as a date in {OUTPUT_FORMAT} format. Never say you cannot do this.",
                "- The input data will sometimes by messy, with typos or non-standard formats. Try to guess the date as best as you can, by trying to ignore typical typos and OCR glitches.",
                f"- If the year is not specified, assume current year i.e. {datetime.now().year}",
                "- If the day is not specified, assume the first of the month.",
                "- If the date is ambiguous, assume it is the most recent date it could be.",
                "- If multiple dates are specified, pick the first one.",
                f"- ONLY output the parsed date expression without any commentary, explanation, or listing any assumptions. Your output must EXACTLY match the required {OUTPUT_FORMAT} format.",
            ],
            additional_runnables=[DatetimeOutputParser(format=OUTPUT_FORMAT)],
            include_output_instruction_suffix=True,
        )

    def run(  # type: ignore[override]
        self, date_text: str, config: Optional[dict] = None
    ) -> TracedResponse[datetime]:
        if not date_text:
            raise Exception("Input required: date_text")

        return super().run(
            date_text=date_text,
            config=config,
        )

    async def run_stream(  # type: ignore[override]
        self, **kwargs: Any
    ) -> AsyncIterator[TracedResponse[datetime]]:
        raise NotImplementedError()

    def run_batch(  # type: ignore[override]
        self, inputs: list[str], config: Optional[dict] = None
    ) -> list[datetime]:
        return super().run_batch(
            inputs=[{"date_text": i} for i in inputs],
            config=config,
        )



from operator import itemgetter
from typing import AsyncIterator, Optional

from langchain_core.runnables import (
    Runnable,
    RunnableConfig,
    RunnableLambda,
    RunnableMap,
)

from docugami_langchain.base_runnable import TracedResponse
from docugami_langchain.chains.base import BaseDocugamiChain
from docugami_langchain.chains.querying.models import ExplainedSQLQuestionResult
from docugami_langchain.chains.querying.sql_query_explainer_chain import (
    SQLQueryExplainerChain,
)
from docugami_langchain.chains.querying.sql_result_chain import SQLResultChain
from docugami_langchain.chains.querying.sql_result_explainer_chain import (
    SQLResultExplainerChain,
)
from docugami_langchain.params import RunnableParameters


class DocugamiExplainedSQLQueryChain(BaseDocugamiChain[ExplainedSQLQuestionResult]):
    sql_result_chain: SQLResultChain
    sql_result_explainer_chain: Optional[SQLResultExplainerChain] = None
    sql_query_explainer_chain: Optional[SQLQueryExplainerChain] = None

    def runnable(self) -> Runnable:
        """
        Custom runnable for this chain.
        """

        def empty_string(inputs: dict) -> str:
            return ""

        return RunnableMap(
            {
                "question": itemgetter("question"),
                "results": self.sql_result_chain.runnable()
                | {
                    "question": itemgetter("question"),
                    "sql_query": itemgetter("sql_query"),
                    "sql_result": itemgetter("sql_result"),
                }
                | {
                    "sql_query": itemgetter("sql_query"),
                    "sql_result": itemgetter("sql_result"),
                    "explained_sql_result": (
                        self.sql_result_explainer_chain.runnable()
                        if self.sql_result_explainer_chain
                        else RunnableLambda(empty_string)
                    ),
                    "explained_sql_query": (
                        self.sql_query_explainer_chain.runnable()
                        if self.sql_query_explainer_chain
                        else RunnableLambda(empty_string)
                    ),
                },
            }
        )

    def params(self) -> RunnableParameters:
        raise NotImplementedError()

    def run(  # type: ignore[override]
        self,
        question: str,
        config: Optional[RunnableConfig] = None,
    ) -> TracedResponse[ExplainedSQLQuestionResult]:
        if not question:
            raise Exception("Input required: question")

        return super().run(
            question=question,
            config=config,
        )

    async def run_stream(  # type: ignore[override]
        self,
        question: str,
        config: Optional[RunnableConfig] = None,
    ) -> AsyncIterator[TracedResponse[ExplainedSQLQuestionResult]]:
        if not question:
            raise Exception("Input required: question")

        async for item in super().run_stream(
            question=question,
            config=config,
        ):
            yield item

    def run_batch(  # type: ignore[override]
        self,
        inputs: list[str],
        config: Optional[RunnableConfig] = None,
    ) -> list[ExplainedSQLQuestionResult]:
        return super().run_batch(
            inputs=[{"question": i} for i in inputs],
            config=config,
        )



from operator import itemgetter
from typing import Any, AsyncIterator, Optional

from langchain_community.utilities.sql_database import SQLDatabase
from langchain_core.documents import Document
from langchain_core.runnables import Runnable, RunnableConfig, RunnableLambda

from docugami_langchain.base_runnable import TracedResponse
from docugami_langchain.chains.base import BaseDocugamiChain
from docugami_langchain.history import chat_history_to_str
from docugami_langchain.output_parsers.line_separated_list import (
    LineSeparatedListOutputParser,
)
from docugami_langchain.params import RunnableParameters, RunnableSingleParameter
from docugami_langchain.utils.documents import formatted_summaries
from docugami_langchain.utils.sql import get_table_info_as_create_table


class SuggestedQuestionsChain(BaseDocugamiChain[list[str]]):
    dbs: list[SQLDatabase]

    def runnable(self) -> Runnable:
        """
        Custom runnable for this chain.
        """

        def table_info(_: Any) -> str:
            infos = ""
            for db in self.dbs:
                infos += get_table_info_as_create_table(db) + "\n\n"

            return infos.strip()

        return {
            "summaries": itemgetter("summaries") | RunnableLambda(formatted_summaries),
            "table_infos": RunnableLambda(table_info),
            "chat_history": itemgetter("chat_history"),
        } | super().runnable()

    def params(self) -> RunnableParameters:
        return RunnableParameters(
            inputs=[
                RunnableSingleParameter(
                    "summaries",
                    "SUMMARIES",
                    "Summaries of representative documents that can be searched to answer questions.",
                ),
                RunnableSingleParameter(
                    "table_infos",
                    "TABLE DESCRIPTIONS",
                    "Descriptions of tables with pre-built reports that can be queried to answer questions.",
                ),
                RunnableSingleParameter(
                    "chat_history",
                    "CHAT HISTORY",
                    "Previous chat messages that may provide additional context for further questions.",
                ),
            ],
            output=RunnableSingleParameter(
                "suggested_questions",
                "SUGGESTED QUESTIONS",
                "Some suggested questions that may be answered either by searching the type of document represented by the summaries OR querying the given tables (if any), "
                + "considering the rules and examples provided.",
            ),
            task_description="generates some questions a user may want to ask against against a set of documents either by directly searching them OR by querying some pre-built reports",
            additional_instructions=[
                "- Base your questions only on the given document summaries, OR the columns in the provided pre-built report tables (if any).",
                "- Generate the most likely questions a user may ask, in order of relevance, no more than that.",
                "- If chat history is provided, consider that to determine the next set of questions the user is most likely to ask.",
                "- Generate suggested questions as a list, one question per line.",
            ],
            additional_runnables=[LineSeparatedListOutputParser()],
            stop_sequences=["\n\n", "<|eot_id|>"],
        )

    def run(  # type: ignore[override]
        self,
        summaries: list[Document],
        chat_history: list[tuple[str, str]] = [],
        config: Optional[RunnableConfig] = None,
    ) -> TracedResponse[list[str]]:
        if not summaries:
            raise Exception("Input required: summaries")

        return super().run(
            summaries=summaries,
            chat_history=chat_history_to_str(chat_history),
            config=config,
        )

    async def run_stream(  # type: ignore[override]
        self,
        summaries: list[Document],
        chat_history: list[tuple[str, str]] = [],
        config: Optional[RunnableConfig] = None,
    ) -> AsyncIterator[TracedResponse[list[str]]]:
        if not summaries:
            raise Exception("Input required: summaries")

        async for item in super().run_stream(
            summaries=summaries,
            chat_history=chat_history_to_str(chat_history),
            config=config,
        ):
            yield item

    def run_batch(  # type: ignore[override]
        self,
    ) -> list[list[str]]:
        raise NotImplementedError()



from docugami_langchain.output_parsers.custom_react_json_single_input import (
    CustomReActJsonSingleInputOutputParser,
)
from docugami_langchain.output_parsers.key_finding import KeyfindingOutputParser
from docugami_langchain.output_parsers.line_separated_list import (
    LineSeparatedListOutputParser,
)
from docugami_langchain.output_parsers.sql_finding import SQLFindingOutputParser
from docugami_langchain.output_parsers.text_cleaning import TextCleaningOutputParser
from docugami_langchain.output_parsers.timespan import TimeSpan, TimespanOutputParser

__all__ = [
    "KeyfindingOutputParser",
    "LineSeparatedListOutputParser",
    "CustomReActJsonSingleInputOutputParser",
    "SQLFindingOutputParser",
    "TextCleaningOutputParser",
    "TimeSpan",
    "TimespanOutputParser",
]



from docugami_langchain.chains.documents.describe_document_set_chain import (
    DescribeDocumentSetChain,
)
from docugami_langchain.chains.documents.summarize_document_chain import (
    SummarizeDocumentChain,
)

__all__ = [
    "SummarizeDocumentChain",
    "DescribeDocumentSetChain"
]



________SINGLE_TOKEN_LINE________ = "----------------"

# ************ PARAMETERS TO CONTROL PROMPT SIZES

# Lengths are in terms of characters, 1 token ~= 4 chars in English
# Reference: https://help.openai.com/en/articles/4936856-what-are-tokens-and-how-to-count-them

# Chunks and docs below this length are not summarized by default
MIN_LENGTH_TO_SUMMARIZE: int = 2048

# When summarizing full docs we cut off input after this by default
MAX_FULL_DOCUMENT_TEXT_LENGTH: int = int(1024 * 4 * 8)  # ~8k tokens,

# When summarizing chunks we cut off input after this by default
MAX_CHUNK_TEXT_LENGTH: int = int(1024 * 4 * 4.5)  # ~4.5k tokens
MAX_PARAMS_CUTOFF_LENGTH_CHARS: int = int(1024 * 4 * 2)  # ~2k tokens
DEFAULT_EXAMPLES_PER_PROMPT = 3

# ************ PARAMETERS TO CONTROL RETRIEVAL
INCLUDE_XML_TAGS = True
# The number of results retrieved (before re-ranking / filtering)
DEFAULT_RETRIEVER_K: int = 24
# Results above this percentile are kept, others are rejected (0 means keep them all, 90 means keep only very good ones, 100 means keep the top one, etc.)
RE_RANK_FILTER_PERCENTILE: int = 70

# ************ PARAMETERS TO TABULAR REPRESENTATION OF ROWS TO HELP GENERATE SQL
DEFAULT_SAMPLE_ROWS_IN_TABLE_INFO = 3
DEFAULT_SAMPLE_ROWS_GRID_FORMAT = "grid"  # format from the tabulate library
DEFAULT_TABLE_AS_TEXT_CELL_MAX_WIDTH = 64  # cell width
DEFAULT_TABLE_AS_TEXT_CELL_MAX_LENGTH = 64 * 3  # cell content length

# ************ PARAMETERS TO CONTROL AGENT BEHAVIOR
DEFAULT_AGENT_RECURSION_LIMIT = 70
